import argparse

import yaml
from yaml.scanner import ScannerError
from importlib.machinery import SourceFileLoader

TYPE = 'type'
LIST = 'list'
DESCRIPTION = 'description'
REQUIRED = 'required'
DEFAULT = 'default'
ALLOWED = 'allowed'
VALUES_DSC = 'values_description'
ONE_OF = 'one of'
SCHEMA = 'schema'
EXAMPLES = 'examples'
ANYOF = 'anyof'
NO_DSC = '(no description)'
VALIDATOR = 'check_with'

NoneType = type(None)


class TextBlock(object):
    def __init__(self, text, tab_replacement='  ', ending=''):
        """

        :type text: str
        """
        self.text = str(text).replace('\t', tab_replacement)
        self.lines = self.text.splitlines()
        self.width = max([len(line) for line in self.lines] + [0])
        self.padded_width = self.width + 2
        self.height = len(self.lines)

    def get_line(self, item, raise_index_error=False, default=''):
        try:
            return self.lines[item]
        except IndexError:
            if raise_index_error:
                raise
            else:
                return default

    def get_line_justified(self, item, fillchar=' ', raise_index_error=False, default=''):
        return self.get_line(item, raise_index_error, default).ljust(self.width, fillchar)

    def __str__(self):
        return self.text


def to_text_block(method):
    def decorated(content):
        if not isinstance(content, TextBlock):
            return method(TextBlock(content))
        else:
            return method(TextBlock)

    return decorated


class RSTRenderer(object):
    def with_escape(method):
        def escaped(content):
            return method(RSTRenderer.escape(content))

        return escaped

    @staticmethod
    def any_of_table(blocks):
        """

        :type blocks: list of TextBlock
        """
        HEADER = 'any of'
        cnt = len(blocks)
        # no need table for single content
        if cnt < 2:
            return blocks[0] if blocks else ''
        # width = widths of contents + separators
        width = max((len(HEADER), sum([c.padded_width for c in blocks]))) + (cnt + 1)
        height = max([c.height for c in blocks])
        # rows separators
        top_bar = '+{}+'.format('-' * (width - 2))
        header_bar = '+{}+'.format('+'.join(['=' * c.padded_width for c in blocks]))
        bottom_bar = '+{}+'.format('+'.join(['-' * c.padded_width for c in blocks]))

        header = '|{}|'.format(HEADER.center(width - 2))
        body = '\n'.join(
            ['| {} |'.format(' | '.join([c.get_line_justified(i) for c in blocks])) for i in range(height)])
        return '\n'.join([top_bar,
                          header,
                          header_bar,
                          body,
                          bottom_bar])

    @staticmethod
    def preserve_indents(block):
        """

        :type block: TextBlock
        """
        return '\n'.join(['| {}'.format(line) for line in block.lines])

    @staticmethod
    def bold(content):
        """
        :type content: str
        :return: str
        """
        return '\n'.join(['**{}**'.format(line) for line in content.splitlines()])

    @staticmethod
    def title(content, new_line_replacement=' ', tab_replacement='  '):
        """
        Underlines content with '='. New lines and tabs will be replaced
        :param str content:
        :param str new_line_replacement:
        :param str tab_replacement:
        :return: str
        """
        prepared_content = content.strip().replace('\n', new_line_replacement).replace('\t', tab_replacement)
        return '{}\n{}'.format(prepared_content, '=' * len(prepared_content))

    @staticmethod
    def subtitle(content, new_line_replacement=' ', tab_replacement='  '):
        prepared_content = content.strip().replace('\n', new_line_replacement).replace('\t', tab_replacement)
        return '{}\n{}'.format(prepared_content, '-' * len(prepared_content))

    @staticmethod
    @with_escape
    @to_text_block
    def italic(block):
        """

        :type block: TextBlock
        """
        return '\n'.join(['*{}*'.format(line) for line in block.lines])

    @staticmethod
    @to_text_block
    def mono(block):
        """

        :type block: TextBlock
        """
        return '\n'.join(['``{}``'.format(line) for line in block.lines])

    @classmethod
    def bullet_list(cls, blocks):
        """

        :type blocks: list of TextBlock
        :rtype: TextBlock
        """
        return TextBlock('\n'.join([cls._list_item(block) for block in blocks]))

    @staticmethod
    def _list_item(block):
        """

        :type block: TextBlock
        """
        return '- ' + '\n  '.join(block.lines)

    @staticmethod
    def def_list(items, sort=True, newlines=True):
        def format_value(value):
            if isinstance(value, (int, bool, NoneType)):
                return format_value(str(value))
            if isinstance(value, str):
                return '\n '.join(value.splitlines())
            elif isinstance(value, TextBlock):
                return '\n '.join(value.lines)
            elif isinstance(value, dict):
                return '\n '.join(RSTRenderer.def_list(value, sort, newlines).splitlines())
            elif isinstance(value, list):
                return '\n '.join(RSTRenderer.bullet_list([TextBlock(item) for item in value]).lines)
            else:
                raise ValueError('Unsupported value type: {}\n{}'.format(type(value), value))

        sort = sorted if sort else lambda x: x
        template = '{}\n {}' if newlines else ':{}: {}'
        return '\n' + '\n'.join([template.format(k.replace('\n', ' '),
                                                 format_value(v).strip())
                                 for k, v in sort(items.items())]) if items else ''

    @staticmethod
    def field_list(items, sort=True, newlines=True):
        """

        :param bool newlines: add newlines between names and values
        :param bool sort: sort items alphabetically by key
        :type items: dict
        :rtype: TextBlock
        """

        def format_value(value):
            if isinstance(value, (int, bool, NoneType)):
                return format_value(str(value))
            if isinstance(value, str):
                return '\n '.join(value.splitlines())
            elif isinstance(value, TextBlock):
                return '\n '.join(value.lines)
            elif isinstance(value, dict):
                return '\n '.join(RSTRenderer.field_list(value, sort, newlines).splitlines())
            elif isinstance(value, list):
                return '\n '.join(RSTRenderer.bullet_list([TextBlock(item) for item in value]).lines)
            else:
                raise ValueError('Unsupported value type: {}\n{}'.format(type(value), value))

        sort = sorted if sort else lambda x: x
        template = ':{}:\n {}' if newlines else ':{}: {}'
        return '\n' + '\n'.join([template.format(k.replace('\n', ' '),
                                                 format_value(v).strip())
                                 for k, v in sort(items.items())]) if items else ''

    @staticmethod
    def dict_list_structure(items, sort_dict=True):
        if isinstance(items, str):
            return TextBlock(items)
        elif isinstance(items, int):
            return TextBlock(str(items))
        elif isinstance(items, list):
            return RSTRenderer.bullet_list([RSTRenderer.dict_list_structure(item) for item in items])
        elif isinstance(items, dict):
            return RSTRenderer.field_list({k: RSTRenderer.dict_list_structure(v) for k, v in items.items()}, sort_dict)

    @staticmethod
    def escape(content):
        """
        :type content: str
        """
        return content.replace('-', r'\-')

    del with_escape


def render_body(renderer, option_kwargs, exclude_keys, special_keys=None):
    """

    :type option_kwargs: dict
    :type exclude_keys: list
    :type special_keys: dict
    """
    common_formatters = {
        EXAMPLES: lambda examples: renderer.def_list({renderer.mono(example): annotation for example, annotation in examples.items()})
    }

    def default_fmt(x):
        return x

    special_keys = special_keys or {}
    special_part = '\n'.join([special_handler(renderer, option_kwargs[special_key])
                              for special_key, special_handler in special_keys.items()
                              if special_key in option_kwargs])
    uncommon_keys = set(exclude_keys) | set(special_keys.keys())
    common_part = renderer.field_list({
        k: common_formatters.get(k, default_fmt)(v)
        for k, v in option_kwargs.items()
        if k not in uncommon_keys
    })

    return '\n'.join([_ for _ in [common_part, special_part] if _])


def render_values_description(renderer, option_kwargs):
    values_description_dict = {
        value: option_kwargs[VALUES_DSC].get(value, '') for value in option_kwargs[ALLOWED]
    } \
        if ALLOWED in option_kwargs \
        else \
        option_kwargs[VALUES_DSC]
    values_description = renderer.field_list(
        {renderer.mono(value): dsc for value, dsc in values_description_dict.items()},
        newlines=False
    )
    return renderer.field_list({ONE_OF: values_description})


def allowed(renderer, values):
    return renderer.field_list({ONE_OF: '[{}]'.format(', '.join([renderer.mono(value) for value in values]))},
                               newlines=False)


class OptionFormatter(object):
    def __init__(self, option_schema):
        """

        :type option_schema: dict
        """
        self.option_name, self.option_kwargs = next(iter(option_schema.items()))
        # print(option_name, option_kwargs)
        self.formatter = self.__guess_formatter()

    def format_dsc(self, renderer):
        dsc = self.option_kwargs.get(DESCRIPTION, NO_DSC).strip('. ')
        if DEFAULT in self.option_kwargs:
            default_value = self.option_kwargs.get(DEFAULT)
            if default_value == '':
                default_value = '""'
            return ' '.join([renderer.italic('- {}. Default:'.format(dsc)),
                             renderer.mono(default_value)])
        elif REQUIRED in self.option_kwargs:
            return renderer.italic('- {}.'.format(dsc)) +\
                ' ' +\
                renderer.bold('Required.')
        else:
            return renderer.italic('- {}.'.format(dsc))

    def scalar_formatter(self, renderer, header=True):
        hdr = renderer.subtitle(renderer.mono(self.option_name) + ' ' + '({})'.format(self.option_kwargs.get(TYPE))) \
            if header else ''
        dsc = self.format_dsc(renderer)
        body = render_body(renderer, self.option_kwargs, [VALIDATOR, TYPE, DESCRIPTION, DEFAULT, REQUIRED], {'allowed': allowed})
        return '\n'.join([_ for _ in [hdr, dsc, body] if _])

    def scalar_with_values_description(self, renderer, header=True):
        hdr = renderer.subtitle(renderer.mono(self.option_name) + ' ' + '({})'.format(self.option_kwargs.get(TYPE))) \
            if header else ''
        dsc = self.format_dsc(renderer)
        body = render_body(renderer, self.option_kwargs, [VALIDATOR, TYPE, DESCRIPTION, DEFAULT, REQUIRED, ALLOWED, VALUES_DSC])
        values_description_block = render_values_description(renderer, self.option_kwargs)
        return '\n'.join([_ for _ in [hdr, dsc, body, values_description_block] if _])

    def dict_formatter(self, renderer, header=True):
        hdr = renderer.subtitle(renderer.mono(self.option_name) + ' ' + '({})'.format(self.option_kwargs.get(TYPE))) \
            if header else ''
        dsc = self.format_dsc(renderer)

        dict_schema = self.option_kwargs[SCHEMA]

        schema_block = renderer.field_list({
            '{} ({})'.format(renderer.mono(key), dict_schema[key].get(TYPE, 'anyof')): get_formatter({key: value})(renderer, header=False)
            for key, value in dict_schema.items()})
        body = render_body(renderer, self.option_kwargs, [VALIDATOR, TYPE, DESCRIPTION, DEFAULT, REQUIRED, SCHEMA])
        return '\n'.join([_ for _ in [hdr, dsc, schema_block, body] if _])

    def anyof_formatter(self, renderer, header=True):
        types = [case[TYPE] for case in self.option_kwargs[ANYOF] if TYPE in case]
        hdr = renderer.subtitle(renderer.mono(self.option_name) + ' ' + '({})'.format(' or '.join(types))) \
            if header else ''
        dsc = self.format_dsc(renderer)
        values_description_block = render_values_description(renderer, self.option_kwargs) \
            if VALUES_DSC in self.option_kwargs else ''
        body = render_body(renderer, self.option_kwargs, [VALIDATOR, TYPE, DESCRIPTION, DEFAULT, REQUIRED, ANYOF, VALUES_DSC])

        return '\n'.join([_ for _ in [hdr, dsc, values_description_block, body] if _])

    def list_formatter(self, renderer, header=True):
        schema = self.option_kwargs[SCHEMA]
        hdr = renderer.subtitle(renderer.mono(self.option_name) + ' '
                                + '({} of {})'.format(self.option_kwargs.get(TYPE, LIST), schema.get(TYPE, '')))
        dsc = self.format_dsc(renderer)
        body = render_body(renderer, self.option_kwargs, [VALIDATOR, TYPE, DEFAULT, REQUIRED, DESCRIPTION, SCHEMA])
        if set(schema.keys()) - {TYPE}:
            schema_block = renderer.field_list({
                '[list_element] ({})'.format(schema.get(TYPE, '')):
                    get_formatter({'list_element': schema})(renderer, header=False)
            })
            return '\n'.join([_ for _ in [hdr, dsc, schema_block, body] if _])
        else:
            return '\n'.join([_ for _ in [hdr, dsc, body] if _])

    def __guess_formatter(self):
        if ANYOF in self.option_kwargs:
            return self.anyof_formatter
        elif SCHEMA in self.option_kwargs:
            return self.list_formatter if self.option_kwargs.get(TYPE) == LIST else self.dict_formatter
        elif VALUES_DSC in self.option_kwargs:
            return self.scalar_with_values_description
        else:
            return self.scalar_formatter


def get_formatter(option_schema):
    """

    :type option_schema: dict
    """
    return OptionFormatter(option_schema).formatter


def format_option(option_schema, renderer):
    return get_formatter(option_schema)(renderer)


def format_schema(schema, renderer, title=None):
    """

    :param dict schema: Cerberus config schema
    :type renderer: RSTRenderer
    """
    body = '\n\n'.join(
        sorted([format_option({option_name: option_schema}, renderer) for option_name, option_schema in schema.items()]))

    if title:
        title = renderer.title(title)
        return title + '\n\n' + body
    else:
        return body


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('schema', help='Path to schema file')
    parser.add_argument('-o', '--output_filename', default='output.rst', help='Name for the output rst document')
    parser.add_argument('--title', default=None, help='Document title')
    parser.add_argument('-a', '--append', action='store_true', help='Don\'t overwrite output file')
    args = parser.parse_args()

    schema_path = args.schema
    output_filename = args.output_filename
    title = args.title
    append = args.append

    try:
        with open(schema_path) as f:
            schema = yaml.load(f, Loader=yaml.FullLoader)
    except ScannerError:
        schema_module = SourceFileLoader('schema', schema_path).load_module()
        schema = schema_module.OPTIONS
    document = format_schema(schema, RSTRenderer(), title)

    if append:
        with open(output_filename, 'a') as f:
            f.write('\n\n')
            f.write(document)
    else:
        with open(output_filename, 'w') as f:
            f.write(document)


if __name__ == '__main__':
    main()



"""Monitoring collector """
import hashlib
import logging
import sys
import time
from ...common.interfaces import MonitoringDataListener

from ..Telegraf.client import SSHClient, LocalhostClient
from ..Telegraf.config import ConfigManager

logger = logging.getLogger(__name__)


class MonitoringCollector(object):
    """Aggregate data from several collectors

    Workflow:
    plugin creates collector
    collector reads monitoring config and creates agents
    collector creates configs for agents (telegraf, shutdown&startup, custom scripts)
    collector installs agents on targets, send configs on targets
    agent starts startups on target, then starts telegraf
    agent reads output of telegraf, consolidates output, caches 5 seconds and then sends output to collector
    collector polls agents for data, decodes known metrics and counts diffs for diff-like metrics
    collector sends data to listeners

    """

    def __init__(self, disguise_hostnames, kill_old):
        self.kill_old = kill_old
        self.disguise_hostnames = disguise_hostnames
        self.config = None
        self.default_target = None
        self.agents = []
        self.agent_sessions = []
        self.listeners = []
        self.first_data_received = False
        self.__collected_data = []
        self.artifact_files = []
        self.load_start_time = None
        self.config_manager = ConfigManager()
        self.old_style_configs = False
        self.ssh_key_path = None
        self.ssh_timeout = 30

    def add_listener(self, obj):
        self.listeners.append(obj)

    def prepare(self):
        """Prepare for monitoring - install agents etc"""

        # Parse config
        agent_configs = []
        if self.config:
            agent_configs = self.config_manager.getconfig(
                self.config, self.default_target, defaults={'ssh_key_path': self.ssh_key_path})

        # Creating agent for hosts
        for config in agent_configs:
            if config['host'] in ['localhost', '127.0.0.1', '::1']:
                client = LocalhostClient(config, self.old_style_configs, kill_old=self.kill_old)
            else:
                client = SSHClient(config, self.old_style_configs, timeout=self.ssh_timeout, kill_old=self.kill_old)

            logger.debug('Installing monitoring agent. Host: %s', client.host)
            agent_config, startup_config, customs_script = client.install()
            if agent_config:
                self.agents.append(client)
                self.artifact_files.append(agent_config)
            if startup_config:
                self.artifact_files.append(startup_config)
            if customs_script:
                self.artifact_files.append(customs_script)

    def start(self):
        """ Start agents

        execute popen of agent.py on target and start output reader thread.
        """
        [agent.start() for agent in self.agents]
        [agent.reader_thread.start() for agent in self.agents]

    def poll(self):
        """ Poll agents for data
        """
        start_time = time.time()
        for agent in self.agents:
            for collect in agent.reader:
                # don't crash if trash or traceback came from agent to stdout
                if not collect:
                    return 0
                for chunk in collect:
                    ts, prepared_results = chunk
                    if not self.first_data_received and prepared_results:
                        self.first_data_received = True
                        logger.info("Monitoring received first data.")
                    if self.load_start_time and int(
                            ts) >= self.load_start_time:
                        ready_to_send = {
                            "timestamp": int(ts),
                            "data": {
                                self.hash_hostname(agent.host): {
                                    "comment": agent.config.comment,
                                    "metrics": prepared_results
                                }
                            }
                        }
                        self.__collected_data.append(ready_to_send)

        logger.debug(
            'Polling/decoding agents data took: %.2fms',
            (time.time() - start_time) * 1000)

        data = self.__collected_data
        self.__collected_data = []
        return data

    def stop(self):
        """Shutdown agents"""
        logger.debug("Uninstalling monitoring agents")
        for agent in self.agents:
            agent._stop_agent()
        for agent in self.agents:
            try:
                log_filename, data_filename = agent.uninstall()
                self.artifact_files.append(log_filename)
                self.artifact_files.append(data_filename)
            except Exception as exc:
                logger.warning("Error while uninstalling agent %s", exc, exc_info=True)
        for agent in self.agents:
            try:
                logger.debug('Waiting for agent %s reader thread to finish.', agent)
                agent.reader_thread.join(10)
                self.agents.remove(agent)
            except BaseException:
                logger.exception('Monitoring reader thread stuck!')

    def get_rest_data(self):
        return self.__collected_data

    def hash_hostname(self, host):
        if self.disguise_hostnames and host:
            return hashlib.md5(host.encode()).hexdigest()
        else:
            return host


class StdOutPrintMon(MonitoringDataListener):
    """Simple listener, writing data to stdout"""

    def __init__(self):
        MonitoringDataListener.__init__(self)

    def monitoring_data(self, data_list):
        [sys.stdout.write(data) for data in data_list]



import re
import json
from datetime import datetime, timezone
from time import time
from queue import Queue

import requests

from logging import getLogger
from yandextank.common.util import observetime
from yandextank.common.monitoring import MonitoringSensorProtocol


LOGGER = getLogger(__name__)


class YCMonitoringSensor(MonitoringSensorProtocol):
    """
    The YCMonitoringSensor class is intended to taking metrics from cloud
    """

    def __init__(
        self,
        panel_name: str,
        api_host: str,
        token: str,
        query: str,
        folder_id: str,
        queue: Queue,
        request_timeout: int,
        priority_labels: list[str],
        ignore_labels: list[str],
    ):
        query, query_folder_id = parse_yc_monitoring_query(query)
        folder_id = folder_id or query_folder_id

        self.panel_name = panel_name
        self.headers = {'Content-type': 'application/json', 'Authorization': f'Bearer {token}'}
        self.params = {'folderId': folder_id}
        self.endpoint = f'https://{api_host}/monitoring/v2/data/read'

        # for some reason YC Monitoring API returns empty response if query contains `folderId=""` param
        # returns no data:    query='"objects_count"{folderId="correctFolderId", service="storage"}'
        # returns valid data: query='"objects_count"{service="storage"}'
        self.data = {
            'query': query,
            'fromTime': datetime.fromtimestamp(time(), tz=timezone.utc).isoformat(),
            'toTime': (datetime.fromtimestamp(time() + 1, tz=timezone.utc)).isoformat(),
            'downsampling': {'gridInterval': 1500},
        }
        self.queue = queue
        self.is_got_some_data_after_parsing = False
        self.request_timeout = request_timeout

        # vector in queue
        self.vector_params = re.findall(r'as_vector\((.*?)\)', query)
        if self.vector_params:
            self.vector_params = self.vector_params[0].split(', ')

        self.priority_labels = priority_labels
        self.ignore_labels = ignore_labels

    def fetch_metrics(self):
        if self.is_got_some_data_after_parsing:
            self.data['fromTime'] = self.data['toTime']
            self.is_got_some_data_after_parsing = False
        self.data['toTime'] = datetime.fromtimestamp(time() + 1, tz=timezone.utc).isoformat()
        self.prepare_data(self.get_data())

    @observetime('YCMonitoringSensor.prepare_data', LOGGER)
    def prepare_data(self, data):
        if isinstance(data, dict) and data.get('metrics'):
            try:
                for idx, sensor_item in enumerate(data['metrics']):
                    if not sensor_item:
                        continue
                    if metrics := self.parse_metrics(sensor_item, idx):
                        self.is_got_some_data_after_parsing = True
                        self.send_metrics(metrics)

            except KeyError as ke:
                LOGGER.warning(f'Wrong data: {data}. {ke}', exc_info=True)
        else:
            LOGGER.warning(f'Wrong data: {data}')

    def parse_metrics(self, metrics: dict, index: int):
        # choosing int or double values
        if metrics['type'] in ['DGAUGE', 'RATE']:
            values_collection = 'doubleValues'
        elif metrics['type'] in ['COUNTER', 'IGAUGE']:
            values_collection = 'int64Values'
        else:
            LOGGER.warning('Unknown metrics type %s', metrics['type'])
            return None

        sensor_name = self.format_sensor(metrics['labels'], index)

        timeseries = metrics.get('timeseries')
        return [
            {
                'sensor': sensor_name,
                'timestamp': tmsp_mill // 1000,
                'value': value,
            }
            for tmsp_mill, value in zip(timeseries['timestamps'], timeseries[values_collection])
            if value != 'NaN'
        ]

    def send_metrics(self, data):
        try:
            self.queue.put(data)
        except (IOError, OSError) as error:
            LOGGER.warning('Sensor %s, send metrics error. %s', data.get('sensor'), error, exc_info=False)

    @observetime('YCMonitoringSensor.get_data', LOGGER)
    def get_data(self, is_first_call: bool = False):
        try:
            response = requests.request(
                'POST',
                self.endpoint,
                headers=self.headers,
                json=self.data,
                params=self.params,
                timeout=self.request_timeout,
            )

            response.raise_for_status()
        except ConnectionError as e:
            if is_first_call:
                raise ConnectionError('After first call got connection error, so you should not add sensors') from e
            LOGGER.exception('Connection error during request')
        except Exception:
            LOGGER.exception('request to YC Cloud Monitoring API failed')
        else:
            try:
                return json.loads(response.text)
            except json.JSONDecodeError as e:
                LOGGER.error('YCMonitoringSensor json parse error: %s', e, exc_info=False)

    def get_sensors(self):
        sensors = set()
        data = self.get_data(is_first_call=True)
        if isinstance(data, dict) and data.get('metrics'):
            for idx, metric in enumerate(data['metrics']):
                sensors.add(self.format_sensor(metric['labels'], idx))
        return sensors

    def format_sensor(self, labels: dict[str, str], index: int) -> str:
        line_name = None
        if not self.vector_params:
            # priority first
            for label in self.priority_labels:
                if label in labels:
                    line_name = labels[label]

            # other labels
            if not line_name:
                parts = []
                for label, value in labels.items():
                    if label in self.ignore_labels:
                        continue
                    else:
                        parts.append(value)
                line_name = '; '.join(parts)
        else:
            line_name = ''.join(['p', self.vector_params[index]])

        return self.panel_name.replace('_', '-') + '_' + line_name


def parse_yc_monitoring_query(query: str) -> tuple[str, str]:
    quote = '("|\')?'
    folderId = '([a-zA-Z0-9._-]+)'
    tail = '\\s*,?\\s*'
    m = re.search(f'folderId={quote}{folderId}{quote}{tail}', query)
    if not m:
        return query, ''

    return query.replace(m.group(0), ''), m.group(2)



import pandas as pd
import time
import itertools as itt
from queue import Empty
from threading import Lock
import threading as th
import logging
logger = logging.getLogger(__name__)


def records_to_df(records):
    records = pd.DataFrame.from_records(records)
    records['receive_ts'] = records['send_ts'] + records['interval_real'] / 1e6
    records['receive_sec'] = records.receive_ts.astype(int)
    # TODO: consider configuration for the following:
    records['tag'] = records.tag.str.rsplit('#', 1, expand=True)[0]
    records.set_index(['receive_sec'], inplace=True)
    return records


def _expand_steps(steps):
    return list(itt.chain(
        * [[rps] * int(duration) for rps, duration in steps]))


class BfgReader(object):
    def __init__(self, results, closed):
        self.buffer = ""
        self.stat_buffer = ""
        self.results = results
        self.closed = closed
        self.records = []
        self.lock = Lock()
        self.thread = th.Thread(target=self._cacher)
        self.thread.start()

    def _cacher(self):
        while True:
            try:
                self.records.append(
                    self.results.get(block=False))
            except Empty:
                if not self.closed.is_set():
                    time.sleep(0.1)
                else:
                    break

    def __next__(self):
        if self.closed.is_set():
            self.thread.join()
            raise StopIteration
        with self.lock:
            records = self.records
            self.records = []
        if records:
            return records_to_df(records)
        return None

    def __iter__(self):
        return self


class BfgStatsReader(object):
    def __init__(self, instance_counter, steps):
        self.closed = False
        self.last_ts = 0
        self.steps = _expand_steps(steps)
        self.instance_counter = instance_counter
        self.start_time = int(time.time())

    def __iter__(self):
        while not self.closed:
            cur_ts = int(time.time())
            if cur_ts > self.last_ts:
                offset = cur_ts - self.start_time
                reqps = 0
                if offset >= 0 and offset < len(self.steps):
                    reqps = self.steps[offset]
                yield [{
                    'ts': cur_ts,
                    'metrics': {
                        'instances': self.instance_counter.value,
                        'reqps': reqps
                    }
                }]
                self.last_ts = cur_ts
            else:
                yield []

    def close(self):
        self.closed = True



'''
Module contains top-level generators.
'''
import hashlib
import json
import logging
import os
import shutil
from builtins import zip
from pathlib import Path

from . import format as fmt
from . import info
from .config import ComponentFactory
from .module_exceptions import DiskLimitError, StepperConfigurationError


class AmmoFactory(object):
    '''
    A generator that produces ammo.
    '''

    def __init__(self, factory):
        '''
        Factory parameter is a configured ComponentFactory that
        is able to produce load plan and ammo generator.
        '''
        self.factory = factory
        self.load_plan = factory.get_load_plan()
        self.ammo_generator = factory.get_ammo_generator()
        self.filter = factory.get_filter()
        self.marker = factory.get_marker()

    def __iter__(self):
        '''
        Returns a generator of (timestamp, marker, missile) tuples
        where missile is in a string representation. Load Plan (timestamps
        generator) and ammo generator are taken from the previously
        configured ComponentFactory, passed as a parameter to the
        __init__ method of this class.
        '''
        ammo_stream = (
            ammo
            for ammo in ((missile, marker or self.marker(missile))
                         for missile, marker in self.ammo_generator)
            if self.filter(ammo))

        return ((timestamp, marker or self.marker(missile), missile)
                for timestamp, (missile, marker
                                ) in zip(self.load_plan, ammo_stream))


class Stepper(object):
    def __init__(self, core, **kwargs):
        info.status = info.StepperStatus()
        info.status.core = core
        self.af = AmmoFactory(ComponentFactory(**kwargs))
        self.ammo = fmt.Stpd(self.af)
        self.first_loop_done = False

    def write(self, f):
        for missile in self.ammo:
            f.write(missile)
            try:
                info.status.inc_ammo_count()
            except StopIteration:
                break

            # check the whole file can be written
            if not self.first_loop_done:
                if info.status.loop_count == 1:
                    # то, что только что закончился цикл чтения фала с патронам,
                    # позволяет нам не заботиться о том, что одни патроны могут быть тяжелее других.
                    self.first_loop_done = True
                    if info.status.max_ammo is None:
                        # Ограничение было задано через опцию loop
                        # Надо выставить ammo_limit вручную,
                        # ибо до прочтения файла непонятно сколько в нём было патронов.
                        if info.status.loop_limit is None:
                            raise StepperConfigurationError('Ammo limit should be specified by '
                                                            'load_type, ammo_limit or loop option. '
                                                            'See https://yandextank.readthedocs.io/en/latest/core_and_modules.html#basic-options')
                        info.status.ammo_limit = info.status.loop_limit * info.status.ammo_count
                        assert info.status.max_ammo is not None
                    self._check_whole_file_can_be_written(f)

    @staticmethod
    def _check_whole_file_can_be_written(file_descriptor):
        if not hasattr(file_descriptor, 'name'):
            # Скорее всего - это IO stream для тестов.
            return
        written_bytes = file_descriptor.tell()
        expected_file_size = (1. / info.status.calculate_lp_progress()) * written_bytes
        need_to_write_more_bytes = expected_file_size - written_bytes
        available_bytes = shutil.disk_usage(Path(file_descriptor.name).parent).free

        reserve = 10 * 000 * 000  # 10 мегабайт
        if available_bytes < need_to_write_more_bytes + reserve:
            raise DiskLimitError('File system has not enough free space for ammo file. '
                                 f'Ammo file expected file size is {expected_file_size} bytes.')


class LoadProfile(object):

    def __init__(self, load_type, schedule):
        self.load_type = load_type
        self.schedule = self.__make_steps(schedule)

    def is_rps(self):
        return self.load_type == 'rps'

    def is_instances(self):
        return self.load_type == 'instances'

    @staticmethod
    def __make_steps(schedule):
        steps = []
        for step in " ".join(schedule.split("\n")).split(')'):
            if step.strip():
                steps.append(step.strip() + ')')
        return steps


class StepperWrapper(object):
    # TODO: review and rewrite this class
    '''
    Wrapper for cached stepper functionality
    '''
    OPTION_LOAD = 'load_profile'
    OPTION_LOAD_TYPE = 'load_type'
    OPTION_SCHEDULE = 'schedule'
    OPTION_STEPS = 'steps'
    OPTION_TEST_DURATION = 'test_duration'
    OPTION_AMMO_COUNT = 'ammo_count'
    OPTION_LOOP = 'loop'
    OPTION_LOOP_COUNT = 'loop_count'
    OPTION_AMMOFILE = "ammofile"
    OPTION_LOADSCHEME = 'loadscheme'
    OPTION_INSTANCES_LIMIT = 'instances'

    def __init__(self, core, cfg):
        self.log = logging.getLogger(__name__)
        self.core = core
        self.cfg = cfg

        self.cache_dir = '.'

        # per-shoot params
        self.instances = 1000
        self.http_ver = '1.0'
        self.ammo_file = None
        self.loop_limit = -1
        self.ammo_limit = -1
        self.uris = []
        self.headers = []
        self.autocases = 0
        self.enum_ammo = False
        self.force_stepping = None
        self.chosen_cases = []

        # out params
        self.stpd = None
        self.steps = []
        self.ammo_count = 1
        self.duration = 0
        self.loop_count = 0
        self.loadscheme = ""
        self.file_cache = 8192

    def get_option(self, option, param2=None):
        ''' get_option wrapper'''
        result = self.cfg[option]
        self.log.debug(
            "Option %s = %s", option, result)
        return result

    @staticmethod
    def get_available_options():
        opts = [
            StepperWrapper.OPTION_AMMOFILE, StepperWrapper.OPTION_LOOP,
            StepperWrapper.OPTION_SCHEDULE, StepperWrapper.OPTION_INSTANCES_LIMIT
        ]
        opts += [
            "instances_schedule", "uris", "headers", "header_http", "autocases",
            "enum_ammo", "ammo_type", "ammo_limit"
        ]
        opts += [
            "use_caching", "cache_dir", "force_stepping", "file_cache",
            "chosen_cases"
        ]
        return opts

    def read_config(self):
        ''' stepper part of reading options '''
        self.log.info("Configuring StepperWrapper...")
        self.ammo_file = self.get_option(self.OPTION_AMMOFILE)
        self.ammo_type = self.get_option('ammo_type')
        if self.ammo_file:
            self.ammo_file = os.path.expanduser(self.ammo_file)
        self.loop_limit = self.get_option(self.OPTION_LOOP)
        self.ammo_limit = self.get_option("ammo_limit")

        self.load_profile = LoadProfile(**self.get_option('load_profile'))

        self.instances = int(
            self.get_option(self.OPTION_INSTANCES_LIMIT, '1000'))
        self.uris = self.get_option("uris", [])
        while '' in self.uris:
            self.uris.remove('')
        self.headers = self.get_option("headers")
        self.http_ver = self.get_option("header_http")
        self.autocases = self.get_option("autocases")
        self.enum_ammo = self.get_option("enum_ammo")
        self.use_caching = self.get_option("use_caching")

        self.file_cache = self.get_option('file_cache')
        cache_dir = self.get_option("cache_dir") or self.core.artifacts_base_dir
        self.cache_dir = os.path.expanduser(cache_dir)
        self.force_stepping = self.get_option("force_stepping")
        if self.get_option(self.OPTION_LOAD)[self.OPTION_LOAD_TYPE] == 'stpd_file':
            self.stpd = self.get_option(self.OPTION_LOAD)[self.OPTION_SCHEDULE]

        self.chosen_cases = self.get_option("chosen_cases").encode('utf8').split()
        if self.chosen_cases:
            self.log.info("chosen_cases LIMITS: %s", self.chosen_cases)

    def prepare_stepper(self):
        ''' Generate test data if necessary '''

        def publish_info(stepper_info):
            info.status.publish('loadscheme', stepper_info.loadscheme)
            info.status.publish('loop_count', stepper_info.loop_count)
            info.status.publish('steps', stepper_info.steps)
            info.status.publish('duration', stepper_info.duration)
            info.status.ammo_count = stepper_info.ammo_count
            info.status.publish('instances', stepper_info.instances)
            self.core.publish('stepper', 'loadscheme', stepper_info.loadscheme)
            self.core.publish('stepper', 'loop_count', stepper_info.loop_count)
            self.core.publish('stepper', 'steps', stepper_info.steps)
            self.core.publish('stepper', 'duration', stepper_info.duration)
            self.core.publish('stepper', 'ammo_count', stepper_info.ammo_count)
            self.core.publish('stepper', 'instances', stepper_info.instances)
            return stepper_info

        if not self.stpd:
            self.stpd = self.__get_stpd_filename()
            if self.use_caching and not self.force_stepping and os.path.exists(
                    self.stpd) and os.path.exists(self.__si_filename()):
                self.log.info("Using cached stpd-file: %s", self.stpd)
                stepper_info = self.__read_cached_options()
                if self.instances and self.load_profile.is_rps():
                    self.log.info(
                        "rps_schedule is set. Overriding cached instances param from config: %s",
                        self.instances)
                    stepper_info = stepper_info._replace(
                        instances=self.instances)
                publish_info(stepper_info)
            else:
                if (
                        self.force_stepping and os.path.exists(self.__si_filename())):
                    os.remove(self.__si_filename())
                self.__make_stpd_file()
                stepper_info = info.status.get_info()
                self.__write_cached_options(stepper_info)
        else:
            self.log.info("Using specified stpd-file: %s", self.stpd)
            stepper_info = publish_info(self.__read_cached_options())
        self.ammo_count = stepper_info.ammo_count
        self.duration = stepper_info.duration
        self.loop_count = stepper_info.loop_count
        self.loadscheme = stepper_info.loadscheme
        self.steps = stepper_info.steps
        if stepper_info.instances:
            self.instances = stepper_info.instances

    def __si_filename(self):
        '''Return name for stepper_info json file'''
        return "%s_si.json" % self.stpd

    def __get_stpd_filename(self):
        ''' Choose the name for stepped data file '''
        if self.use_caching:
            sep = "|"
            hasher = hashlib.md5()
            hashed_str = "cache version 6" + sep + \
                ';'.join(self.load_profile.schedule) + sep + str(self.loop_limit)
            hashed_str += sep + str(self.ammo_limit) + sep + ';'.join(
                self.load_profile.schedule) + sep + str(self.autocases)
            hashed_str += sep + ";".join(self.uris) + sep + ";".join(
                self.headers) + sep + self.http_ver + sep + b';'.join(
                    self.chosen_cases).decode('utf8')
            hashed_str += sep + str(self.enum_ammo) + sep + str(self.ammo_type)
            if self.load_profile.is_instances():
                hashed_str += sep + str(self.instances)
            if self.ammo_file:
                opener = self.core.resource_manager.get_opener(self.ammo_file)
                hashed_str += sep + opener.hash
            else:
                if not self.uris:
                    raise RuntimeError("Neither ammofile nor uris specified")
                hashed_str += sep + \
                    ';'.join(self.uris) + sep + ';'.join(self.headers)
            self.log.debug("stpd-hash source: %s", hashed_str)
            hasher.update(hashed_str.encode('utf8'))
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir)
            stpd = self.cache_dir + '/' + \
                os.path.basename(self.ammo_file) + \
                "_" + hasher.hexdigest() + ".stpd"
        else:
            stpd = os.path.realpath("ammo.stpd")
        self.log.debug("Generated cache file name: %s", stpd)
        return stpd

    def __read_cached_options(self):
        '''
        Read stepper info from json
        '''
        self.log.debug("Reading cached stepper info: %s", self.__si_filename())
        with open(self.__si_filename(), 'r') as si_file:
            si = info.StepperInfo(**json.load(si_file))
        return si

    def __write_cached_options(self, si):
        '''
        Write stepper info to json
        '''
        self.log.debug("Saving stepper info: %s", self.__si_filename())
        with open(self.__si_filename(), 'w') as si_file:
            json.dump(si._asdict(), si_file, indent=4)

    def __make_stpd_file(self):
        ''' stpd generation using Stepper class '''
        self.log.info("Making stpd-file: %s", self.stpd)
        stepper = Stepper(
            self.core,
            rps_schedule=self.load_profile.schedule if self.load_profile.is_rps() else None,
            http_ver=self.http_ver,
            ammo_file=self.ammo_file,
            instances_schedule=self.load_profile.schedule if self.load_profile.is_instances() else None,
            instances=self.instances,
            loop_limit=self.loop_limit,
            ammo_limit=self.ammo_limit,
            uris=self.uris,
            headers=[header.strip('[]') for header in self.headers],
            autocases=self.autocases,
            enum_ammo=self.enum_ammo,
            ammo_type=self.ammo_type,
            chosen_cases=self.chosen_cases,
            use_cache=self.use_caching,
            resource_manager=self.core.resource_manager,)
        with open(self.stpd, 'wb', self.file_cache) as os:
            stepper.write(os)



from setuptools import setup, find_packages
from yandextank.version import VERSION


setup(
    package_dir={"": "."},
    name='yandextank',
    version=VERSION,
    description='a performance measurement tool',
    longer_description='''
Yandex.Tank is a performance measurement and load testing automatization tool.
It uses other load generators such as JMeter, ab or phantom inside of it for
load generation and provides a common configuration system for them and
analytic tools for the results they produce.
''',
    python_requires='>=3.8',
    maintainer='Yandex Load Team',
    maintainer_email='load@yandex-team.ru',
    url='http://yandex.github.io/yandex-tank/',
    packages=find_packages(exclude=["tests", "tmp", "docs", "data"]),
    install_requires=[
        'psutil>=5.9.8', 'requests>=2.31.0',
        'pandas<=2.0.3', 'numpy<=1.26.4',
        'pip>=24.0',
        'pyyaml>=5.4', 'cerberus>=1.3.5', 'influxdb>=5.3.1',
        'retrying>=1.3.4', 'pytest-runner', 'pyserial', 'retrying', 'six>=1.16.0',
        'environ-config>=23.2.0',
    ],
    setup_requires=[],
    tests_require=[
        'pytest>=7.4.4', 'flake8', 'pytest-benchmark', 'zipp==0.5.1', 'mock'
    ],
    license='LGPLv2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Testing :: Traffic Generation',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={
        'console_scripts': [
            'yandex-tank = yandextank.core.cli:main',
            'tank-postloader = yandextank.plugins.DataUploader.cli:post_loader',
            'tank-docs-gen = yandextank.validator.docs_gen:main'
        ],
    },
    package_data={
        'yandextank.api': ['config/*'],
        'yandextank.core': ['config/*'],
        'yandextank.aggregator': ['config/*'],
        'yandextank.plugins.Android': ['binary/*', 'config/*'],
        'yandextank.plugins.Autostop': ['config/*'],
        'yandextank.plugins.Bfg': ['config/*'],
        'yandextank.plugins.Console': ['config/*'],
        'yandextank.plugins.DataUploader': ['config/*'],
        'yandextank.plugins.InfluxUploader': ['config/*'],
        'yandextank.plugins.OpenTSDBUploader': ['config/*'],
        'yandextank.plugins.JMeter': ['config/*'],
        'yandextank.plugins.JsonReport': ['config/*'],
        'yandextank.plugins.Pandora': ['config/*'],
        'yandextank.plugins.Phantom': ['config/*'],
        'yandextank.plugins.RCAssert': ['config/*'],
        'yandextank.plugins.ResourceCheck': ['config/*'],
        'yandextank.plugins.ShellExec': ['config/*'],
        'yandextank.plugins.ShootExec': ['config/*'],
        'yandextank.plugins.Telegraf': ['config/*', 'agent/*'],
        'yandextank.plugins.NeUploader': ['config/*'],
        'yandextank.plugins.YCMonitoring': ['config/*'],
    },)



# TODO: merge with Yandex Tank validator
import yaml
import pkg_resources
import logging
import importlib.util

from cerberus import Validator
from yandextank.contrib.netort.netort.data_manager.common.util import recursive_dict_update


logger = logging.getLogger(__name__)


class ValidationError(Exception):
    pass


def load_yaml_schema(path):
    with open(path) as f:
        return yaml.load(f)


def load_py_schema(path):
    spec = importlib.util.spec_from_file_location('schema', path)
    schema_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(schema_module)
    return schema_module.SCHEMA


def load_schema(directory):
    try:
        return load_yaml_schema(directory)
    except IOError:
        try:
            return load_py_schema(directory)
        except ImportError:
            raise IOError('Neither .yaml nor .py schema found in %s' % directory)


class ValidatedConfig(object):
    def __init__(
            self,
            configs,
            dynamic_options,
            package_schema_path,
            package_schema_file='config/schema.yaml',
            with_dynamic_options=True,
            core_section='core'
    ):
        if not isinstance(configs, list):
            configs = [configs]
        self.__raw_config_dict = self.__load_multiple(
            [config for config in configs if config is not None])
        if self.__raw_config_dict.get(core_section) is None:
            self.__raw_config_dict[core_section] = {}
        self.BASE_SCHEMA = load_yaml_schema(
            pkg_resources.resource_filename(
                package_schema_path, package_schema_file
            )
        )
        self.DYNAMIC_OPTIONS = dynamic_options
        self.CORE_SECTION = core_section
        self._validated = None
        self.with_dynamic_options = with_dynamic_options
        logger.debug('patched_raw_config_dict: %s', self.__raw_config_dict)

    def __load_multiple(self, configs):
        configs_count = len(configs)
        if configs_count == 0:
            return {}
        elif configs_count == 1:
            return configs[0]
        elif configs_count == 2:
            return recursive_dict_update(configs[0], configs[1])
        else:
            return self.__load_multiple(
                [recursive_dict_update(configs[0], configs[1])] + configs[2:])

    def get_option(self, section, option, default=None):
        try:
            self.validated[section][option]
        except KeyError:
            if default is not None:
                return default
            else:
                raise KeyError()
        return self.validated[section][option]

    def get_enabled_sections(self):
        return [
            section_name for section_name, section_config in self.__raw_config_dict.items()
            if section_config.get('enabled', False)
        ]

    def has_option(self, section, option):
        return self.validated

    @property
    def validated(self):
        if not self._validated:
            self._validated = self.__validate()
        return self._validated

    def save(self, filename):
        with open(filename, 'w') as f:
            yaml.dump(self.validated, f)

    def __validate(self):
        core_validated = self.__validate_core()
        errors = {}
        if len(errors) > 0:
            raise ValidationError(dict(errors))
        return core_validated

    def __validate_core(self):
        v = Validator(self.BASE_SCHEMA)
        result = v.validate(self.__raw_config_dict, self.BASE_SCHEMA)
        if not result:
            raise ValidationError(v.errors)
        normalized = v.normalized(self.__raw_config_dict)
        return self.__set_core_dynamic_options(normalized) if self.with_dynamic_options else normalized

    def __set_core_dynamic_options(self, config):
        for option, setter in self.DYNAMIC_OPTIONS.items():
            config[self.CORE_SECTION][option] = setter()
        return config



# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
import json
import logging
import os
import sys
import pandas as pd
from datetime import timedelta, datetime

import io

from collections import defaultdict
from ...common.interfaces import AbstractPlugin, AggregateResultListener

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def calc_overall_times(overall, quantiles):
    cumulative = overall.cumsum()
    total = cumulative.max()
    positions = cumulative.searchsorted([float(i) / 100 * total for i in quantiles])
    all_times = [cumulative.index[i] / 1000. for i in positions]
    overall_times = zip(quantiles, all_times)
    return overall_times


def calc_duration(first_ts, last_ts):
    first_time = timedelta(seconds=first_ts)
    last_time = timedelta(seconds=last_ts)
    return str(last_time - first_time + timedelta(seconds=1))


def make_resp_json(overall_times, overall_proto_code, overall_net_code, duration, loadscheme, time_start,
                   autostop_info):
    quant = {}
    for q, t in overall_times:
        quant['q' + str(q)] = t
    res = {
        "duration": duration,
        "time_start": time_start,
        "loadscheme": loadscheme,
        "quantiles": quant,
        "proto_code": overall_proto_code,
        "net_code": overall_net_code
    }

    if autostop_info:
        res['autostop_rps'] = autostop_info['rps']
        res['autostop_reason'] = autostop_info['reason']

    try:
        response = json.dumps(res, indent=2, sort_keys=False)
    except ValueError as e:
        logger.warning('Can\'t convert offline report to json format: %s', e, exc_info=True)
        response = None
    return response


def make_resp_text(overall_times, overall_proto_code, overall_net_code, duration, loadscheme, time_start,
                   autostop_info):
    res = ['Duration: {:>8}\n'.format(duration)]
    res.append('Loadscheme: {}\n'.format(loadscheme))
    if autostop_info:
        res.append('Autostop rps: {}\n'.format(autostop_info['rps']))
        res.append('Autostop reason: {}\n'.format(autostop_info['reason']))
    res.append('Start time: {}\n'.format(time_start))
    res.append('Percentiles all ms:\n')
    for q, t in overall_times:
        res.append('{:>5}% < {:>5}\n'.format(q, t))
    res.append('HTTP codes(code/count):\n')
    for q, t in overall_proto_code.items():
        res.append('{:>5}: {:<7}\n'.format(q, t))
    res.append('Net codes(code/count):\n')
    for q, t in overall_net_code.items():
        res.append('{:>5}: {:<7}\n'.format(q, t))
    return ''.join(res)


class Plugin(AbstractPlugin, AggregateResultListener):
    # pylint:disable=R0902
    SECTION = 'offline_report'

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        try:
            self.data_and_stats_stream = io.open(os.path.join(self.core.artifacts_dir,
                                                              self.get_option('offline_data_log')),
                                                 mode='w')
            self.add_cleanup(lambda: self.data_and_stats_stream.close())
            self.overall_json_stream = io.open(os.path.join(self.core.artifacts_dir,
                                                            self.get_option('offline_json_report')),
                                               mode='w')
            self.add_cleanup(lambda: self.overall_json_stream.close())
            self.overall_text_stream = io.open(os.path.join(self.core.artifacts_dir,
                                                            self.get_option('offline_text_report')),
                                               mode='w')
            self.add_cleanup(lambda: self.overall_text_stream.close())
        except Exception:
            logging.exception('Failed to open file')
            raise OSError('Error opening OfflineReport log file')

        self.overall = None
        self.overall_proto_code = defaultdict(int)
        self.overall_net_code = defaultdict(int)
        self.quantiles = [10, 20, 30, 40, 50, 60, 70, 75, 80, 85, 90, 95, 99, 99.5, 100]
        self.print_report = self.get_option("print_report")
        self.first_ts = None
        self.last_ts = None
        self.stats = None

    def get_available_options(self):
        return ['offline_data_log', 'offline_json_report', 'offline_text_report', 'print_report']

    def configure(self):
        self.core.job.subscribe_plugin(self)

    def prepare_test(self):
        self.data_and_stats_stream.write('[')

    def on_aggregated_data(self, data, stats):
        """
        @data: aggregated data
        @stats: stats about gun
        """

        last_proto_code = data['overall']['proto_code']['count']
        for code, count in last_proto_code.items():
            self.overall_proto_code[code] += count

        last_net_code = data['overall']['net_code']['count']
        for code, count in last_net_code.items():
            self.overall_net_code[code] += count

        self.data_and_stats_stream.write(
            '%s,\n' % json.dumps({
                'ts': stats['ts'],
                'instances': stats['metrics']['instances'],
                'reqps': stats['metrics']['reqps'],
                'quantiles': {
                    "q50": int(data['overall']['interval_real']['q']['value'][0]),
                    "q75": int(data['overall']['interval_real']['q']['value'][1]),
                    "q80": int(data['overall']['interval_real']['q']['value'][2]),
                    "q85": int(data['overall']['interval_real']['q']['value'][3]),
                    "q90": int(data['overall']['interval_real']['q']['value'][4]),
                    "q95": int(data['overall']['interval_real']['q']['value'][5]),
                    "q98": int(data['overall']['interval_real']['q']['value'][6]),
                    "q99": int(data['overall']['interval_real']['q']['value'][7]),
                    "q100": int(data['overall']['interval_real']['q']['value'][8]),
                },
                'proto_code': last_proto_code,
                'net_code': data['overall']['net_code']['count']
            }))

        incoming_hist = data['overall']['interval_real']['hist']
        dist = pd.Series(incoming_hist['data'], index=incoming_hist['bins'])
        if self.overall is None:
            self.overall = dist
        else:
            self.overall = self.overall.add(dist, fill_value=0)

        if self.first_ts is None:
            self.first_ts = stats['ts']
        else:
            self.last_ts = stats['ts']

    def post_process(self, retcode):
        try:
            self.data_and_stats_stream.seek(self.data_and_stats_stream.tell() - 2, os.SEEK_SET)
            self.data_and_stats_stream.write(']')
        except ValueError as e:
            logger.error('Can\'t write offline report %s', e)

        overall_times = calc_overall_times(self.overall, self.quantiles)
        stepper_info = self.core.info.get_value(['stepper'])

        try:
            duration = str(timedelta(seconds=stepper_info['duration']))
        except (KeyError, TypeError) as e:
            logger.error('Can\'t get test duration %s', e)
            duration = calc_duration(self.first_ts, self.last_ts)

        try:
            loadscheme = ' '.join(stepper_info['loadscheme'])
        except (KeyError, TypeError) as e:
            logger.error('Can\'t get test loadscheme %s', e)
            loadscheme = None

        generator_info = self.core.info.get_value(['generator'])
        try:
            time_start = datetime.fromtimestamp(generator_info['time_start']).strftime("%Y-%m-%d %H:%M:%S")
        except (KeyError, TypeError) as e:
            logger.error('Can\'t get test start time %s', e)
            time_start = datetime.fromtimestamp(self.first_ts).strftime("%Y-%m-%d %H:%M:%S")

        try:
            autostop_info = self.core.info.get_value(['autostop'])
        except (KeyError, TypeError) as e:
            logger.error('Can\'t get autostop info %s', e)
            autostop_info = None

        resp_json = make_resp_json(
            overall_times,
            self.overall_proto_code,
            self.overall_net_code,
            duration,
            loadscheme,
            time_start,
            autostop_info
        )
        if resp_json is not None:
            self.overall_json_stream.write('%s' % resp_json)

        resp_text = make_resp_text(
            overall_times,
            self.overall_proto_code,
            self.overall_net_code,
            duration,
            loadscheme,
            time_start,
            autostop_info
        )
        self.overall_text_stream.write('%s' % resp_text)

        if self.print_report:
            sys.stdout.write(resp_text)

        return retcode



''' Plugin provides fullscreen console '''
import logging
import sys
import threading
import traceback

from .screen import Screen
from ...common.interfaces import AbstractPlugin, AggregateResultListener


class Plugin(AbstractPlugin, AggregateResultListener):
    ''' Console plugin '''
    SECTION = 'console'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.log = logging.getLogger(__name__)
        self.screen = None
        self.render_exception = None
        self.console_markup = None
        self.info_panel_width = self.get_option("info_panel_width")
        self.short_only = self.get_option("short_only")
        # these three provide non-blocking console output
        self.__console_view = None
        self.__writer_thread = None
        self.__writer_event = None
        self.cases_sort_by = self.get_option("cases_sort_by")
        self.cases_max_spark = self.get_option("cases_max_spark")
        self.max_case_len = self.get_option("max_case_len")
        self.times_max_spark = self.get_option("times_max_spark")
        self.sizes_max_spark = self.get_option("sizes_max_spark")

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return [
            "info_panel_width", "short_only", "disable_all_colors",
            "disable_colors"
        ]

    def configure(self):
        if not self.get_option("disable_all_colors"):
            self.console_markup = RealConsoleMarkup()
        else:
            self.console_markup = NoConsoleMarkup()
        for color in self.get_option("disable_colors").split(' '):
            self.console_markup.__dict__[color] = ''
        self.screen = Screen(
            self.info_panel_width, self.console_markup,
            cases_sort_by=self.cases_sort_by,
            cases_max_spark=self.cases_max_spark,
            max_case_len=self.max_case_len,
            times_max_spark=self.times_max_spark,
            sizes_max_spark=self.sizes_max_spark
        )
        try:
            aggregator = self.core.job.aggregator
            aggregator.add_result_listener(self)
        except KeyError:
            self.log.debug("No aggregator for console")
            self.screen.block_rows = []
            self.screen.info_panel_percent = 100

    def __console_writer(self):
        while True:
            self.__writer_event.wait()
            self.__writer_event.clear()

            if self.__console_view:
                if not self.short_only:
                    self.log.debug("Writing console view to STDOUT")
                    sys.stdout.write(self.console_markup.clear)
                    sys.stdout.write(self.__console_view.decode('utf8'))
                    sys.stdout.write(self.console_markup.TOTAL_RESET)

    def is_test_finished(self):
        if not self.__writer_thread:
            self.__writer_event = threading.Event()
            self.__writer_thread = threading.Thread(
                target=self.__console_writer)
            self.__writer_thread.daemon = True
            self.__writer_thread.start()

        try:
            self.__console_view = self.screen.render_screen().encode('utf-8')
        except Exception as ex:
            self.log.warning("Exception inside render: %s", traceback.format_exc())
            self.render_exception = ex
            self.__console_view = ""

        self.__writer_event.set()
        return -1

    def on_aggregated_data(self, data, stats):
        # TODO: use stats data somehow
        if self.short_only:
            overall = data.get('overall')

            quantiles = dict(
                zip(
                    overall['interval_real']['q']['q'],
                    overall['interval_real']['q']['value']))
            info = (
                "ts:{ts}\tRPS:{rps}\tavg:{avg_rt:.2f}\t"
                "min:{min:.2f}\tmax:{q100:.2f}\tq95:{q95:.2f}\t").format(
                    ts=data.get('ts'),
                    rps=overall['interval_real']['len'],
                    avg_rt=float(overall['interval_real']['total'])
                    / overall['interval_real']['len'] / 1000.0,
                    min=overall['interval_real']['min'] / 1000.0,
                    q100=quantiles[100] / 1000,
                    q95=quantiles[95] / 1000)
            self.log.info(info)
        else:
            self.screen.add_second_data(data)

    def add_info_widget(self, widget):
        ''' add right panel widget '''
        if not self.screen:
            self.log.debug("No screen instance to add widget")
        else:
            self.screen.add_info_widget(widget)


# ======================================================


class RealConsoleMarkup(object):
    '''
    Took colors from here: https://www.siafoo.net/snippet/88
    '''
    WHITE_ON_BLACK = '\033[37;40m'
    TOTAL_RESET = '\033[0m'
    clear = "\x1b[2J\x1b[H"
    new_line = "\n"

    YELLOW = '\033[1;33m'
    RED = '\033[1;31m'
    RED_DARK = '\033[31;3m'
    RESET = WHITE_ON_BLACK + '\033[1;m'
    CYAN = "\033[1;36m"
    GREEN = "\033[1;32m"
    WHITE = "\033[1;37m"
    MAGENTA = '\033[1;35m'
    BG_MAGENTA = '\033[1;45m'
    BG_GREEN = '\033[1;42m'
    BG_BROWN = '\033[1;43m'
    BG_CYAN = '\033[1;46m'

    def get_markup_vars(self):
        return [
            self.YELLOW, self.RED, self.RESET, self.CYAN, self.BG_MAGENTA,
            self.WHITE, self.BG_GREEN, self.GREEN, self.BG_BROWN,
            self.RED_DARK, self.MAGENTA, self.BG_CYAN
        ]

    def clean_markup(self, orig_str):
        ''' clean markup from string '''
        for val in self.get_markup_vars():
            orig_str = orig_str.replace(val, '')
        return orig_str


# ======================================================
# FIXME: 3 better way to have it?


class NoConsoleMarkup(RealConsoleMarkup):
    ''' all colors are disabled '''
    WHITE_ON_BLACK = ''
    TOTAL_RESET = ''
    clear = ""
    new_line = "\n"

    YELLOW = ''
    RED = ''
    RED_DARK = ''
    RESET = ''
    CYAN = ""
    GREEN = ""
    WHITE = ""
    MAGENTA = ''
    BG_MAGENTA = ''
    BG_GREEN = ''
    BG_BROWN = ''
    BG_CYAN = ''


# ======================================================



class PluginImplementationError(RuntimeError):
    """
    Error in plugin implementation
    """
    pass


class PluginNotPrepared(Exception):
    """
    Can't find plugin's info in core.job
    """

    def __init__(self, msg=None):
        self.message = "%s\n%s" % (self.__doc__, msg)


class GeneratorNotFound(Exception):
    """
    Can't find generator in config
    """
    pass



''' Pandora config generator '''
import json
from pkg_resources import resource_string

from ...stepper.util import parse_duration


def periodic_schedule(batch_size, period, limit):
    return {
        "LimiterType": "periodic",
        "Parameters": {
            "BatchSize": float(batch_size),
            "MaxCount": float(limit),
            "Period": float(period),
        },
    }


def linear_schedule(start_rps, end_rps, period):
    return {
        "LimiterType": "linear",
        "Parameters": {
            "StartRps": float(start_rps),
            "EndRps": float(end_rps),
            "Period": parse_duration(period) / 1000.0,
        },
    }


def unlimited_schedule(*args):
    return {
        "LimiterType": "unlimited",
        "Parameters": {},
    }


step_producers = {
    "periodic": periodic_schedule,
    "linear": linear_schedule,
    "unlimited": unlimited_schedule,
}


def parse_schedule(schedule):
    steps = [
        step.strip() for step in " ".join(schedule.split("\n")).split(')')
        if step.strip()
    ]
    if len(steps) > 1:
        raise NotImplementedError("Composite schedules not implemented yet")
    schedule_type, params = steps[0].split('(')
    params = [p.strip() for p in params.split(',')]
    if schedule_type in step_producers:
        return step_producers[schedule_type](*params)
    else:
        raise NotImplementedError("Step of type %s is not implemented" % schedule_type)


class PandoraConfig(object):
    def __init__(self):
        self.pools = []

    def data(self):
        return {"Pools": [p.data() for p in self.pools]}

    def json(self):
        return json.dumps(self.data(), indent=2)

    def add_pool(self, pool_config):
        self.pools.append(pool_config)


class PoolConfig(object):
    def __init__(self):
        self.config = json.loads(
            resource_string(__name__, 'config/pandora_pool_default.json'))

    def set_ammo(self, ammo):
        self.config["AmmoProvider"]["AmmoSource"] = ammo

    def set_ammo_type(self, ammo_type):
        self.config["AmmoProvider"]["AmmoType"] = ammo_type

    def set_loop(self, loop):
        self.config["AmmoProvider"]["Passes"] = loop

    def set_report_file(self, report_file):
        self.config["ResultListener"]["Destination"] = report_file

    def set_startup_schedule(self, startup_schedule):
        self.config["StartupLimiter"] = startup_schedule

    def set_user_schedule(self, user_schedule):
        self.config["UserLimiter"] = user_schedule

    def set_shared_schedule(self, shared_schedule):
        self.config["SharedSchedule"] = shared_schedule

    def set_target(self, target):
        self.config["Gun"]["Parameters"]["Target"] = target

    def set_ssl(self, ssl):
        self.config["Gun"]["Parameters"]["SSL"] = ssl

    def set_gun_type(self, gun_type):
        self.config["Gun"]["GunType"] = gun_type

    def data(self):
        return self.config


def main():
    pool_config = PoolConfig()
    pool_config.set_loop(1)
    pool_config.set_startup_schedule(parse_schedule("periodic(100, 100, 100)"))
    pool_config.set_target("example.org:443")
    pandora_config = PandoraConfig()
    pandora_config.add_pool(pool_config)
    print(pandora_config.json())


if __name__ == '__main__':
    main()



import logging
import time
from collections import namedtuple
from sys import stdout

log = logging.getLogger(__name__)

StepperInfo = namedtuple('StepperInfo', 'loop_count,steps,loadscheme,duration,ammo_count,instances')


class LoopCountLimit(Exception):
    pass


class StepperStatus(object):
    '''
    Raises StopIteration when limits are reached.
    '''

    def __init__(self):
        self.core = None  # dirty hack. StepperWrapper should pass core here.
        self.info = {
            'loop_count': 0,
            'steps': None,
            'loadscheme': None,
            'duration': None,
            'ammo_count': 0,
            'instances': None,
        }
        self._ammo_count = 0
        self._old_ammo_count = 0
        self._loop_count = 0
        self._af_position = None
        self.af_size = None
        self.loop_limit = None
        self.ammo_limit = None
        self.lp_len = None
        self.lp_progress = 0
        self.af_progress = 0
        self._timer = time.time()

    def publish(self, key, value):
        if key not in self.info:
            raise RuntimeError("Tried to publish to a non-existent key: %s" % key)
        log.debug('Published %s to %s', value, key)
        self.info[key] = value

    @property
    def af_position(self):
        return self._af_position

    @af_position.setter
    def af_position(self, value):
        self._af_position = value
        self.update_af_progress()

    @property
    def ammo_count(self):
        return self._ammo_count

    @ammo_count.setter
    def ammo_count(self, value):
        self._ammo_count = value
        self.update_lp_progress()
        if self.ammo_limit and value > self.ammo_limit:
            print()
            log.info("Ammo limit reached: %s", self.ammo_limit)
            raise StopIteration

    def inc_ammo_count(self):
        self.ammo_count += 1

    @property
    def loop_count(self):
        return self._loop_count

    @loop_count.setter
    def loop_count(self, value):
        self._loop_count = value
        if self.loop_limit and value >= self.loop_limit:
            print()  # do not overwrite status (go to new line)
            log.warning("Ammo loop limit reached: %s", self.loop_limit)
            raise LoopCountLimit

    def inc_loop_count(self):
        self.loop_count += 1

    def get_info(self):
        self.info['ammo_count'] = self._ammo_count
        self.info['loop_count'] = self._loop_count
        for key in self.info:
            if self.info[key] is None:
                raise RuntimeError("Information for %s is not published yet." % key)
        return StepperInfo(**self.info)

    def update_view(self):
        ammo_generated = self._ammo_count - self._old_ammo_count
        self._old_ammo_count = self._ammo_count
        cur_time = time.time()
        time_delta = cur_time - self._timer
        self._timer = cur_time
        if time_delta > 0:
            stdout.write(
                "AF: %3s%%, LP: %3s%%, loops: %10s, speed: %5s Krps\r" % (
                    self.af_progress, self.lp_progress, self.loop_count,
                    int(ammo_generated / time_delta / 1000.0)))
            stdout.flush()
            if self.core:
                self.core.publish("stepper", "progress", self.lp_progress)
                self.core.publish("stepper", "loop_count", self.loop_count)
                self.core.publish(
                    "stepper", "speed",
                    "%s Krps" % int(ammo_generated / time_delta / 1000.0))

    def update_af_progress(self):
        if self.af_size and self.loop_limit and self.af_position is not None:
            bytes_read = self.af_size * self.loop_count + self.af_position
            total_bytes = self.af_size * self.loop_limit
            progress = int(float(bytes_read) / float(total_bytes) * 100.0)
        else:
            progress = 100
        if self.af_progress != progress:
            self.af_progress = progress
            self.update_view()

    @property
    def max_ammo(self):
        if self.ammo_limit is not None and self.lp_len is not None:
            return min(self.ammo_limit, self.lp_len)
        return self.ammo_limit or self.lp_len

    def calculate_lp_progress(self):
        if self.max_ammo is None:
            return 1.
        return float(self.ammo_count) / float(self.max_ammo)

    def update_lp_progress(self):
        progress = int(self.calculate_lp_progress() * 100.0)
        if self.lp_progress != progress:
            self.lp_progress = progress
            self.update_view()


status = StepperStatus()



OPTIONS = {
    "additional_libs": {
        "type": "list",
        "default": [],
        'description': 'Libs for Phantom, to be added to phantom config file in section "module_setup"',
        'schema': {
            'type': 'string'
        }
    },
    "address": {
        'description': 'Address of target. Format: [host]:port, [ipv4]:port, [ipv6]:port. Port is optional. '
                       'Tank checks each test if port is available',
        "type": "string",
        "empty": False,
        "required": True,
        'examples': {'127.0.0.1:8080': '', 'www.w3c.org': ''}
    },
    'autocases': {
        'description': 'Use to automatically tag requests. Requests might be grouped by tag for later analysis.',
        'anyof': [
            {'type': 'integer'},
            {'type': 'string',
             'allowed': ['uri', 'uniq']}],
        'default': 0,
        'values_description': {
            'uri': 'tag each request with its uri path, slashes are replaced with underscores',
            'uniq': 'tag each request with unique uid',
            '<N>': 'use N first uri parts to tag request, slashes are replaced with underscores'
        },
        'examples': {
            2: '/example/'
               'search/hello/help/us?param1=50 -> _example_search',
            3: '/example/search/hello/help/us?param1=50 -> _example_search_hello',
            'uri': '/example/search/hello/help/us?param1=50 -> _example_search_hello_help_us',
            'uniq': '/example/search/hello/help/us?param1=50 -> c98b0520bb6a451c8bc924ed1fd72553'
        }
    },
    "affinity": {
        'description': 'Use to set CPU affinity',
        "type": "string",
        "default": '',
        'examples': {
            '0-3': 'enable first 4 cores',
            '0,1,2,16,17,18': 'enable 6 specified cores'
        }
    },
    'ammo_limit': {
        'description': 'Sets the upper limit for the total number of requests',
        'type': 'integer',
        'default': -1
    },
    'ammo_type': {
        'description': 'Ammo format. Don\'t forget to change ammo_type option if you switch the format of your ammo, otherwise you might get errors',
        'type': 'string',
        'default': 'phantom',
        'allowed': ['phantom', 'uri', 'uripost', 'access'],
        'values_description': {
            'phantom': 'Use Request-style file. Most versatile, HTTP as is. See tutorial for details',
            'uri': 'Use URIs listed in file with headers. Simple but allows for GET requests only. See tutorial for details',
            'uripost': 'Use URI-POST file. Allows POST requests with bodies. See tutorial for details',
            'access': 'Use access.log from your web server as a source of requests'
        },
        'tutorial_link': 'http://yandextank.readthedocs.io/en/latest/tutorial.html#preparing-requests'
    },
    'ammofile': {
        'type': 'string',
        'default': '',
        'description': 'Path to ammo file. Ammo file contains requests to be sent to a server. Can be gzipped',
        'tutorial_link': 'http://yandextank.readthedocs.io/en/latest/tutorial.html#preparing-requests',
    },
    "buffered_seconds": {
        "type": "integer",
        "default": 2,
        'description': 'Aggregator latency'
    },
    'cache_dir': {
        'type': 'string',
        'nullable': True,
        'default': None,
        'description': 'stpd-file cache directory'
    },
    'chosen_cases': {
        'type': 'string',
        'default': '',
        'description': 'Use only selected cases.'
    },
    'client_certificate': {
        'type': 'string',
        'default': '',
        'description': 'Path to client SSL certificate'
    },
    'client_cipher_suites': {
        'type': 'string',
        'default': '',
        'description': 'Cipher list, consists of one or more cipher strings separated by colons (see man ciphers)',
    },
    'client_key': {
        'type': 'string',
        'default': '',
        'description': 'Path to client\'s certificate\'s private key'
    },
    'config': {
        'type': 'string',
        'default': '',
        'description': 'Use ready phantom config instead of generated'
    },
    'connection_test': {
        'type': 'boolean',
        'default': True,
        'description': 'Test TCP socket connection before starting the test'
    },
    "enum_ammo": {
        "type": "boolean",
        "default": False
    },
    'file_cache': {
        'type': 'integer',
        'default': 8192
    },
    'force_stepping': {
        'type': 'integer',
        'default': 0,
        'description': 'Ignore cached stpd files, force stepping'
    },
    'gatling_ip': {
        'type': 'string',
        'default': ''
    },
    "header_http": {
        "type": "string",
        'default': '1.0',
        'description': 'HTTP version',
        'allowed': ['1.0', '1.1'],
        'values_description': {
            '1.0': 'http 1.0',
            '1.1': 'http 1.1'
        }
    },
    "headers": {
        "type": "list",
        'default': [],
        'description': 'HTTP headers',
        'schema': {
            'description': 'Format: "Header: Value"',
            'type': 'string',
            'examples': {'accept: text/html': ''}
        }
    },
    'instances': {
        'description': 'Max number of concurrent clients.',
        'type': 'integer',
        'default': 1000
    },
    'loop': {
        'description': 'Loop over ammo file for the given amount of times.',
        'type': 'integer',
        'default': -1
    },
    'method_options': {
        'description': 'Additional options for method objects. It is used for Elliptics etc.',
        'type': 'string',
        'default': ''
    },
    'method_prefix': {
        'description': 'Object\'s type, that has a functionality to create test requests.',
        'type': 'string',
        'default': 'method_stream'
    },
    'multi': {
        'type': 'list',
        'schema': {'type': 'dict'},
        'default': [],
        'description': 'List of configs for multi-test. All of the options from main config supported. All of them not required and inherited from main config if not specified'
    },
    'name': {
        'description': 'Name of a part in multi config',
        'type': 'string',
        'required': False
    },
    'phantom_http_entity': {
        'type': 'string',
        'default': '8M',
        'description': 'Limits the amount of bytes Phantom reads from response.'
    },
    'phantom_http_field': {
        'type': 'string',
        'default': '8K',
        'description': 'Header size.'
    },
    'phantom_http_field_num': {
        'type': 'integer',
        'default': 128,
        'description': 'Max number of headers'
    },
    'phantom_http_line': {
        'type': 'string',
        'default': '1K',
        'description': 'First line length'
    },
    "phantom_modules_path": {
        "type": "string",
        "default": "/usr/lib/phantom",
        'description': 'Phantom modules path.'
    },
    "phantom_path": {
        'description': 'Path to Phantom binary',
        "type": "string",
        "default": "phantom"
    },
    "phout_file": {
        "type": "string",
        'description': 'deprecated',
        "default": ""
    },
    'port': {
        'description': 'Explicit target port, overwrites port defined with address',
        'type': 'string',
        'default': '',
        'regex': r'\d{0,5}'
    },
    "load_profile": {
        'description': 'Configure your load setting the number of RPS or instances (clients) as a function of time,'
                       'or using a prearranged schedule',
        "type": "dict",
        'tutorial_link': 'http://yandextank.readthedocs.io/en/latest/tutorial.html#tutorials',
        'schema': {
            'load_type': {
                'required': True,
                'description': 'Choose control parameter',
                'type': 'string',
                'allowed': ['rps', 'instances', 'stpd_file'],
                'values_description': {
                    'rps': 'control the rps rate',
                    'instances': 'control the number of instances',
                    'stpd_file': 'use prearranged schedule file'}
            },
            'schedule': {
                'type': 'string',
                'required': True,
                'description': 'load schedule or path to stpd file',
                'examples': {
                    'line(100,200,10m)': 'linear growth from 100 to 200 instances/rps during 10 minutes',
                    'const(200,90s)': 'constant load of 200 instances/rps during 90s',
                    'test_dir/test_backend.stpd': 'path to ready schedule file'},
                'check_with': 'load_scheme'
            }
        },
        'required': True
    },
    'source_log_prefix': {
        'description': 'Prefix added to class name that reads source data',
        'type': 'string',
        'default': ''
    },
    'ssl': {
        'description': 'Enable ssl',
        'type': 'boolean',
        'default': False
    },
    "threads": {
        'description': 'Phantom thread count. When not specified, defaults to <processor cores count> / 2 + 1',
        "type": "integer",
        "default": None,
        "nullable": True
    },
    'tank_type': {
        'description': 'Choose between http and pure tcp guns',
        'type': 'string',
        'default': 'http',
        'allowed': ['http', 'none'],
        'values_description': {
            'http': 'HTTP gun',
            'none': 'TCP gun'
        }
    },
    "timeout": {
        'description': 'Response timeout',
        "type": "string",
        "default": "11s"
    },
    "uris": {
        "type": "list",
        'default': [],
        'description': 'URI list',
        'schema': {
            'type': 'string',
            'description': 'URI path string'
        },
        'examples': {
            '["/example/search", "/example/search/hello", "/example/search/hello/help"]': ''
        }
    },
    'use_caching': {
        'description': 'Enable stpd-file caching for similar tests. Set false to reload ammo file and generate new stpd',
        'type': 'boolean',
        'default': True
    },
    "writelog": {
        'description': 'Enable verbose request/response logging.',
        "type": "string",
        "default": "0",
        'allowed': ['0', 'all', 'proto_warning', 'proto_error'],
        'values_description': {
            '0': 'disable',
            'all': 'all messages',
            'proto_warning': '4xx+5xx+network errors',
            'proto_error': '5xx+network errors',
        }
    }
}


MULTI_OPTIONS = {n: {k: v for k, v in d.items() if k != 'required' and k != 'default'} for n, d in OPTIONS.items()}


MULTI = {
    'multi': {
        'type': 'list',
        'allow_unknown': True,
        'schema': {'type': 'dict', 'schema': MULTI_OPTIONS},
        'default': []}
}


def compile_schema():
    schema = OPTIONS.copy()
    schema.update(MULTI)
    return schema


SCHEMA = compile_schema()



import time


def uts(dt):
    return int(time.mktime(dt.timetuple()))


class Decoder(object):
    """
    Decode metrics incoming from tank into points for InfluxDB client

    Parameters
    ----------
    parent_tags : dict
        common per-test tags
    tank_tag : str
        tank identifier tag
    uuid : str
        test id tag
    labeled : bool
        detailed stats for each label
    histograms : bool
        response time histograms measurements
    """
    def __init__(self, tank_tag, uuid, parent_tags, labeled, histograms):
        self.labeled = labeled
        initial_tags = {
            "tank": tank_tag,
            "uuid": uuid
        }
        initial_tags.update(parent_tags)
        self.tags = initial_tags
        self.histograms = histograms

    def set_uuid(self, id_):
        self.tags['uuid'] = id_

    def decode_monitoring(self, data):
        """
        The reason why we have two separate methods for monitoring
        and aggregates is a strong difference in incoming data.
        """
        points = list()
        for second_data in data:
            for host, host_data in second_data["data"].items():
                points.append(
                    self.__make_points(
                        "monitoring",
                        {"host": host, "comment": host_data.get("comment")},
                        second_data["timestamp"],
                        {
                            # cast int to float. avoid https://github.com/yandex/yandex-tank/issues/776
                            metric: float(value) if isinstance(value, int) else value
                            for metric, value in host_data["metrics"].items()
                        }
                    )
                )
        return points

    def decode_aggregates(self, aggregated_data, gun_stats, prefix):
        ts = aggregated_data["ts"]
        points = list()
        # stats overall w/ __OVERALL__ label
        points += self.__make_points_for_label(
            ts,
            aggregated_data["overall"],
            "__OVERALL__",
            prefix,
            gun_stats
        )
        # detailed stats per tag
        if self.labeled:
            for label, aggregated_data_by_tag in aggregated_data["tagged"].items():
                points += self.__make_points_for_label(
                    ts,
                    aggregated_data_by_tag,
                    label,
                    prefix,
                    gun_stats
                )
        return points

    def __make_points_for_label(self, ts, data, label, prefix, gun_stats):
        """x
        Make a set of points for `this` label

        overall_quantiles, overall_meta, net_codes, proto_codes, histograms
        """
        label_points = list()

        label_points.extend(
            (
                # overall quantiles for label
                self.__make_points(
                    prefix + "overall_quantiles",
                    {"label": label},
                    ts,
                    self.__make_quantile_fields(data)
                ),
                # overall meta (gun status) for label
                self.__make_points(
                    prefix + "overall_meta",
                    {"label": label},
                    ts,
                    self.__make_overall_meta_fields(data, gun_stats)
                ),
                # net codes for label
                self.__make_points(
                    prefix + "net_codes",
                    {"label": label},
                    ts,
                    self.__make_netcodes_fields(data)
                ),
                # proto codes for label
                self.__make_points(
                    prefix + "proto_codes",
                    {"label": label},
                    ts,
                    self.__make_protocodes_fields(data)
                )
            )
        )
        # histograms, one row for each bin
        if self.histograms:
            for bin_, count in zip(data["interval_real"]["hist"]["bins"],
                                   data["interval_real"]["hist"]["data"]):
                label_points.append(
                    self.__make_points(
                        prefix + "histograms",
                        {"label": label},
                        ts,
                        {"bin": bin_, "count": count}
                    )
                )
        return label_points

    @staticmethod
    def __make_quantile_fields(data):
        return {
            'q' + str(q): value / 1000.0
            for q, value in zip(data["interval_real"]["q"]["q"],
                                data["interval_real"]["q"]["value"])
        }

    @staticmethod
    def __make_overall_meta_fields(data, stats):
        return {
            "active_threads": stats["metrics"]["instances"],
            "RPS": data["interval_real"]["len"],
            "planned_requests": float(stats["metrics"]["reqps"]),
        }

    @staticmethod
    def __make_netcodes_fields(data):
        return {
            str(code): int(cnt)
            for code, cnt in data["net_code"]["count"].items()
        }

    @staticmethod
    def __make_protocodes_fields(data):
        return {
            str(code): int(cnt)
            for code, cnt in data["proto_code"]["count"].items()
        }

    def __make_points(self, measurement, additional_tags, ts, fields):
        """
        Parameters
        ----------
        measurement : string
            measurement type (e.g. monitoring, overall_meta, net_codes, proto_codes, overall_quantiles)
        additional_tags : dict
            custom additional tags for this points
        ts : integer
            timestamp
        fields : dict
            influxdb columns

        Returns
        -------
        dict
            points for InfluxDB client
        """
        tags = self.tags.copy()
        tags.update(additional_tags)
        return {
            "measurement": measurement,
            "tags": tags,
            "time": int(ts),
            "fields": fields,
        }



from ..common.interfaces import AbstractMetric, TypeEvents, TypeHistogram
import numpy as np


class Event(AbstractMetric):
    def __init__(self, **kw):
        super(Event, self).__init__(**kw)
        self.dtypes = {
            'ts': np.int64,
            'value': np.str,
        }
        self.columns = ['ts', 'value']

    @property
    def type(self):
        return TypeEvents

    @property
    def aggregate_types(self):
        return [TypeHistogram]



import importlib.util
import logging
import time
from contextlib import contextmanager
from random import randint

import requests
from queue import Full
from ...common.interfaces import AbstractPlugin

logger = logging.getLogger(__name__)

requests_logger = logging.getLogger('requests')
requests_logger.setLevel(logging.WARNING)

requests.packages.urllib3.disable_warnings()


class GunConfigError(Exception):
    pass


class AbstractGun(AbstractPlugin):
    def __init__(self, core, cfg):
        super(AbstractGun, self).__init__(core, cfg, 'bfg_gun')
        self.results = None

    @contextmanager
    def measure(self, marker):
        start_time = time.time()
        data_item = {
            "send_ts": start_time,
            "tag": marker,
            "interval_real": None,
            "connect_time": 0,
            "send_time": 0,
            "latency": 0,
            "receive_time": 0,
            "interval_event": 0,
            "size_out": 0,
            "size_in": 0,
            "net_code": 0,
            "proto_code": 200,
        }
        try:
            yield data_item
        except Exception as e:
            logger.warning("%s failed while measuring with %s", marker, e)
            if data_item["proto_code"] == 200:
                data_item["proto_code"] = 500
            if data_item["net_code"] == 0:
                data_item["net_code"] == 1
            raise
        finally:
            if data_item.get("interval_real") is None:
                data_item["interval_real"] = int(
                    (time.time() - start_time) * 1e6)
            try:
                self.results.put(data_item, block=False)
            except Full:
                logger.error("Results full. Data corrupted")

    def setup(self):
        pass

    def shoot(self, missile, marker):
        raise NotImplementedError("Gun should implement 'shoot(self, missile, marker)' method")

    def teardown(self):
        pass

    def get_option(self, key, default_value=None):
        try:
            return super(AbstractGun, self).get_option(key, default_value)
        except KeyError:
            if default_value is not None:
                return default_value
            else:
                raise GunConfigError('Missing key: %s' % key)


class LogGun(AbstractGun):
    SECTION = 'log_gun'

    def __init__(self, core, cfg):
        super(LogGun, self).__init__(core, cfg)
        param = self.get_option("param")
        logger.info('Initialized log gun for BFG with param = %s' % param)

    def shoot(self, missile, marker):
        logger.info("Missile: %s\n%s", marker, missile)
        rt = randint(2, 30000) * 1000
        with self.measure(marker) as di:
            di["interval_real"] = rt


class HttpGun(AbstractGun):
    SECTION = 'http_gun'

    def __init__(self, core, cfg):
        super(HttpGun, self).__init__(core, cfg)
        self.base_address = cfg["base_address"]

    def shoot(self, missile, marker):
        logger.debug("Missile: %s\n%s", marker, missile)
        logger.debug("Sending request: %s", self.base_address + missile)
        with self.measure(marker) as di:
            try:
                r = requests.get(self.base_address + missile, verify=False)
                di["proto_code"] = r.status_code
            except requests.ConnectionError:
                logger.debug("Connection error", exc_info=True)
                di["net_code"] = 1
                di["proto_code"] = 500


class SqlGun(AbstractGun):
    SECTION = 'sql_gun'

    def __init__(self, core):
        super(SqlGun, self).__init__(core)

        from sqlalchemy import create_engine
        from sqlalchemy import exc
        self.exc = exc

        self.engine = create_engine(self.get_option("db"))

    def shoot(self, missile, marker):
        logger.debug("Missile: %s\n%s", marker, missile)
        with self.measure(marker) as di:
            errno = 0
            proto_code = 200
            try:
                cursor = self.engine.execute(missile.replace('%', '%%'))
                cursor.fetchall()
                cursor.close()
            except self.exc.TimeoutError as e:
                logger.debug("Timeout: %s", e)
                errno = 110
            except self.exc.ResourceClosedError as e:
                logger.debug(e)
            except self.exc.SQLAlchemyError as e:
                proto_code = 500
                logger.debug(e.orig.args)
            except self.exc.SAWarning as e:
                proto_code = 400
                logger.debug(e)
            except Exception as e:
                proto_code = 500
                logger.debug(e)
            di["proto_code"] = proto_code
            di["net_code"] = errno


class CustomGun(AbstractGun):
    """
    This gun is deprecated! Use UltimateGun
    """
    SECTION = 'custom_gun'

    def __init__(self, core, cfg):
        super(CustomGun, self).__init__(core, cfg)
        logger.warning("Custom gun is deprecated. Use Ultimate gun instead")
        module_path = cfg["module_path"].strip()
        module_name = cfg["module_name"]
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)

    def shoot(self, missile, marker):
        try:
            self.module.shoot(missile, marker, self.measure)
        except Exception as e:
            logger.warning("CustomGun %s failed with %s", marker, e)

    def setup(self):
        if hasattr(self.module, 'init'):
            self.module.init(self)


class ScenarioGun(AbstractGun):
    """
    This gun is deprecated! Use UltimateGun
    """
    SECTION = 'scenario_gun'

    def __init__(self, core, cfg):
        super(ScenarioGun, self).__init__(core, cfg)
        logger.warning("Scenario gun is deprecated. Use Ultimate gun instead")
        module_path = cfg["module_path"]
        if module_path:
            module_path = module_path.strip()
        else:
            module_path = None
        module_name = cfg["module_name"]
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        self.scenarios = self.module.SCENARIOS

    def shoot(self, missile, marker):
        marker = marker.rsplit("#", 1)[0]  # support enum_ammo
        if not marker:
            marker = "default"
        scenario = self.scenarios.get(marker, None)
        if scenario:
            try:
                scenario(missile, marker, self.measure)
            except Exception as e:
                logger.warning("Scenario %s failed with %s", marker, e)
        else:
            logger.warning("Scenario not found: %s", marker)

    def setup(self):
        if hasattr(self.module, 'init'):
            self.module.init(self)


class UltimateGun(AbstractGun):
    SECTION = "ultimate_gun"

    def __init__(self, core, cfg):
        super(UltimateGun, self).__init__(core, cfg)
        class_name = self.get_option("class_name")
        module_path = self.get_option("module_path")
        if module_path:
            module_path = module_path.strip()
        else:
            module_path = None
        module_name = self.get_option("module_name")
        self.init_param = self.get_option("init_param")
        #
        # Dirty Hack
        #
        # we will add current unix timestamp to the name of a module each time
        # it is imported to be sure Python won't be able to cache it
        #
        hack_module_name = "%s_%d" % (module_name, time.time())
        spec = importlib.util.spec_from_file_location(hack_module_name, module_path)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        test_class = getattr(self.module, class_name, None)
        if not isinstance(test_class, type):
            raise NotImplementedError(
                "Class definition for '%s' was not found in '%s' module" %
                (class_name, module_name))
        self.load_test = test_class(self)

    def setup(self):
        if callable(getattr(self.load_test, "setup", None)):
            self.load_test.setup(self.init_param)

    def teardown(self):
        if callable(getattr(self.load_test, "teardown", None)):
            self.load_test.teardown()

    def shoot(self, missile, marker):
        marker = marker.rsplit("#", 1)[0]  # support enum_ammo
        if not marker:
            marker = "default"
        scenario = getattr(self.load_test, marker, None)
        if callable(scenario):
            try:
                scenario(missile)
            except Exception as e:
                logger.warning("Scenario %s failed with %s", marker, e, exc_info=True)
        else:
            logger.warning("Scenario not found: %s", marker)
